// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// from gst-gir-files (https://gitlab.freedesktop.org/gstreamer/gir-files-rs.git)
// DO NOT EDIT
#![allow(deprecated)]

use crate::ffi;
use glib::{
    prelude::*,
    signal::{SignalHandlerId, connect_raw},
    translate::*,
};
use std::boxed::Box as Box_;

glib::wrapper! {
    #[doc(alias = "GstVideoSink")]
    pub struct VideoSink(Object<ffi::GstVideoSink, ffi::GstVideoSinkClass>) @extends gst_base::BaseSink, gst::Element, gst::Object;

    match fn {
        type_ => || ffi::gst_video_sink_get_type(),
    }
}

impl VideoSink {
    pub const NONE: Option<&'static VideoSink> = None;
}

unsafe impl Send for VideoSink {}
unsafe impl Sync for VideoSink {}

pub trait VideoSinkExt: IsA<VideoSink> + 'static {
    #[doc(alias = "show-preroll-frame")]
    fn shows_preroll_frame(&self) -> bool {
        ObjectExt::property(self.as_ref(), "show-preroll-frame")
    }

    #[doc(alias = "show-preroll-frame")]
    fn set_show_preroll_frame(&self, show_preroll_frame: bool) {
        ObjectExt::set_property(self.as_ref(), "show-preroll-frame", show_preroll_frame)
    }

    #[doc(alias = "show-preroll-frame")]
    fn connect_show_preroll_frame_notify<F: Fn(&Self) + Send + Sync + 'static>(
        &self,
        f: F,
    ) -> SignalHandlerId {
        unsafe extern "C" fn notify_show_preroll_frame_trampoline<
            P: IsA<VideoSink>,
            F: Fn(&P) + Send + Sync + 'static,
        >(
            this: *mut ffi::GstVideoSink,
            _param_spec: glib::ffi::gpointer,
            f: glib::ffi::gpointer,
        ) {
            unsafe {
                let f: &F = &*(f as *const F);
                f(VideoSink::from_glib_borrow(this).unsafe_cast_ref())
            }
        }
        unsafe {
            let f: Box_<F> = Box_::new(f);
            connect_raw(
                self.as_ptr() as *mut _,
                c"notify::show-preroll-frame".as_ptr(),
                Some(std::mem::transmute::<*const (), unsafe extern "C" fn()>(
                    notify_show_preroll_frame_trampoline::<Self, F> as *const (),
                )),
                Box_::into_raw(f),
            )
        }
    }
}

impl<O: IsA<VideoSink>> VideoSinkExt for O {}
